import pytest
from wandb.summary import Summary
from click.testing import CliRunner
import json


@pytest.fixture
def summary():
    with CliRunner().isolated_filesystem():
        s = Summary()
        s.update({"foo": "init"})
        yield s


def test_set_attrs(summary):
    summary.foo = "bar"
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}


def test_get_attr(summary):
    assert summary.foo == "init"


def test_update(summary):
    summary.update({"foo": "bar"})
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}


def test_set_item(summary):
    summary["foo"] = "bar"
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}


def test_get_item(summary):
    assert summary["foo"] == "init"


def test_delete(summary):
    summary.update({"foo": "bar", "bad": True})
    del summary["bad"]
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}



import pytest
import os
from click.testing import CliRunner

from wandb import wandb_config as config
import yaml
import sys
import os
import argparse
import textwrap


def test_config_empty_by_default():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        assert list(conf.keys()) == []


def test_config_empty_by_default():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf['a'] = 15
        conf.b = 16
        assert list(conf.keys()) == ['a', 'b']
        assert conf.a == 15
        assert conf['b'] == 16


def test_config_accepts_dict_vals():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf['a'] = {'b': 14, 'c': 15}
        assert conf.a == {'b': 14, 'c': 15}


def test_config_update_accepts_dict_vals():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf.update({'b': 14, 'c': 15})
        assert conf.b == 14
        assert conf.c == 15


def test_config_persists():
    with CliRunner().isolated_filesystem():
        conf = config.Config(run_dir='.')
        conf['a'] = 5
        conf.b = 14.3

        conf2 = config.Config(config_paths=['config.yaml'])
        assert conf2['a'] == 5
        assert conf2.b == 14.3


def test_config_defaults():
    with CliRunner().isolated_filesystem():
        open('config-defaults.yaml', 'w').write(textwrap.dedent("""\
            wandb_version: 1

            a:
              desc: the number nineteen
              value: 19
            b:
              desc: null
              value: a_cow
            c: 141912
            """))
        conf = config.Config(wandb_dir='.')
        assert conf.a == 19
        assert conf.b == 'a_cow'
        assert conf.c == 141912


def test_config_file_overrides():
    with CliRunner().isolated_filesystem():
        open('config-defaults.yaml', 'w').write(textwrap.dedent("""\
            wandb_version: 1

            a:
              desc: the number nineteen
              value: 19
            b:
              desc: null
              value: a_cow
            """))
        open('config-special.yaml', 'w').write(textwrap.dedent("""\
            wandb_version: 1

            a:
              desc: the number nineteen
              value: 43
            """))
        conf = config.Config(wandb_dir='.', config_paths=[
                             'config-special.yaml'])
        assert conf.a == 43
        assert conf.b == 'a_cow'



import tensorflow as tf
from collections import defaultdict, OrderedDict
import json


def read_tensorflow_log_file(event_file_path):
    scalars = defaultdict(OrderedDict)
    first = None
    for e in tf.train.summary_iterator(event_file_path):
        if not first:
            first = e.wall_time
        for v in e.summary.value:
            if v.simple_value:
                scalars[int(e.wall_time)][v.tag] = v.simple_value

    with open("wandb-history.jsonl", "w") as f:
        for key, value in scalars.items():
            value["_runtime"] = key - first
            value["_time"] = key
            f.write(json.dumps(value) + "\n")

    return scalars



from wandb import wandb_dir
import glob
import os
import json
import re
import logging
import sys
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from wandb.board.app.models import Dir, Settings, RunMutator
from wandb.board.app.util.errors import NotFoundError
from wandb import Error

base_path = wandb_dir()
data = {
    'Runs': []
}
settings = Settings(base_path)


def load(path_override=None):
    global data
    data['Runs'] = []
    root = os.path.abspath(path_override or base_path)
    if os.path.exists(root):
        print("Loading %s ..." % root)
    else:
        raise Error("Directory does not exist: %s" % root)
    settings.path = root
    for path in sorted(glob.glob(root + "/*run-*"), key=lambda p: p.split("run-")[1], reverse=True):
        run_dir = Dir(path)
        data['Runs'].append(run_dir.load())
    watch_dir(root)


def watch_dir(path):
    def on_file_created(event):
        try:
            run_dir = Dir(event.src_path)
        except ValueError:
            return None
        if os.path.isdir(event.src_path):
            # TODO: ensure this is the top level dir?
            if run_dir.run_id and not find_run(run_dir.run_id):
                print("New run started at %s" % run_dir.path)
                data["Runs"].insert(0, run_dir.load())
        run = find_run(run_dir.run_id)
        run_dir.load(run)

    def on_file_modified(event):
        try:
            run_dir = Dir(event.src_path)
        except ValueError:
            return None
        run = find_run(run_dir.run_id)
        run_dir.load(run)

    handler = PatternMatchingEventHandler(
        patterns=[os.path.join(path, "*")], ignore_patterns=['*/.*', '*.tmp'])
    handler.on_created = on_file_created
    handler.on_modified = on_file_modified
    observer = Observer()
    observer.schedule(handler, path, recursive=True)
    observer.start()


def find_run(name, mutator=False):
    if name == "latest":
        return data["Runs"][0]
    else:
        try:
            run = next(run for run in data["Runs"] if run.id == name)
        except StopIteration:
            run = None
        if mutator and run:
            return RunMutator(run)
        elif run is None:
            raise NotFoundError("Run %s not found" % name)
        else:
            return run



import json
import os

import wandb
from wandb import util
from wandb.meta import Meta
import six

SUMMARY_FNAME = 'wandb-summary.json'


class Summary(object):
    """Used to store summary metrics during and after a run."""

    def __init__(self, out_dir='.'):
        self._fname = os.path.join(out_dir, SUMMARY_FNAME)
        self.load()

    def load(self):
        try:
            self._summary = json.load(open(self._fname))
        except (IOError, ValueError):
            self._summary = {}

    def _write(self):
        with open(self._fname, 'w') as f:
            s = util.json_dumps_safer(self._summary, indent=4)
            f.write(s)
            f.write('\n')

    def __getitem__(self, k):
        return self._summary[k]

    def __setitem__(self, k, v):
        self._summary[k] = v
        self._write()

    def __setattr__(self, k, v):
        if k.startswith("_"):
            super(Summary, self).__setattr__(k, v)
        else:
            self._summary[k] = v
            self._write()

    def __getattr__(self, k):
        if k.startswith("_"):
            super(Summary, self).__getattr__(k)
        else:
            return self._summary[k]

    def __delitem__(self, k):
        del self._summary[k]
        self._write()

    def __repr__(self):
        return json.dumps(self._summary, indent=4)

    def get(self, k, default=None):
        return self._summary.get(k, default)

    def update(self, key_vals):
        if not isinstance(key_vals, dict):
            raise wandb.Error('summary.update expects dict')
        # TODO: This removes media from the summary, but will silently remove a user provided dict with _type
        self._summary.update(
            {k: v for k, v in six.iteritems(key_vals) if not (isinstance(v, dict) and v.get("_type"))})
        self._write()



import datetime
import logging
import os
import random
import time
import traceback

import wandb
from wandb import util

logger = logging.getLogger(__name__)


def make_printer(msg):
    def printer():
        print(msg)
    return printer


class Retry(object):
    """Creates a retryable version of a function.

    Calling this will call the passed function, retrying if any exceptions in
    retryable_exceptions are caught, with exponential backoff.
    """

    MAX_SLEEP_SECONDS = 5 * 60

    def __init__(self, call_fn, retry_timedelta=None, num_retries=None,
                 retryable_exceptions=None, error_prefix="network error"):
        self._call_fn = call_fn
        self._error_prefix = error_prefix
        self._retry_timedelta = retry_timedelta
        self._num_retries = num_retries
        self._retryable_exceptions = retryable_exceptions
        if self._retryable_exceptions is None:
            self._retryable_exceptions = (Exception,)
        self._index = 0

    @property
    def num_iters(self):
        """The number of iterations the previous __call__ retried."""
        return self._num_iter

    def __call__(self, *args, **kwargs):
        """Call the wrapped function, with retries.

        Args:
           retry_timedelta (kwarg): amount of time to retry before giving up.
           sleep_base (kwarg): amount of time to sleep upon first failure, all other sleeps
               are derived from this one.
        """

        retry_timedelta = kwargs.pop('retry_timedelta', self._retry_timedelta)
        if retry_timedelta is None:
            retry_timedelta = datetime.timedelta(days=1000000)

        num_retries = kwargs.pop('num_retries', self._num_retries)
        if num_retries is None:
            num_retries = 1000000

        if os.environ.get('WANDB_TEST'):
            num_retries = 0

        sleep_base = kwargs.pop('retry_sleep_base', 1)

        first = True
        sleep = sleep_base
        start_time = datetime.datetime.now()
        now = start_time

        self._num_iter = 0

        while True:
            try:
                result = self._call_fn(*args, **kwargs)
                if not first:
                    wandb.termlog('%s resolved after %s, resuming normal operation.' % (
                        self._error_prefix, datetime.datetime.now() - start_time))
                return result
            except self._retryable_exceptions as e:
                if (datetime.datetime.now() - start_time >= retry_timedelta
                        or self._num_iter >= num_retries):
                    raise
                if self._num_iter == 2:
                    logger.exception('Retry attempt failed:')
                    wandb.termlog(
                        '%s (%s), entering retry loop. See %s for full traceback.' % (
                            self._error_prefix, e.__class__.__name__, util.get_log_file_path()))
                if os.getenv('WANDB_DEBUG'):
                    traceback.print_exc()
            first = False
            time.sleep(sleep + random.random() * 0.25 * sleep)
            sleep *= 2
            if sleep > self.MAX_SLEEP_SECONDS:
                sleep = self.MAX_SLEEP_SECONDS
            now = datetime.datetime.now()

            self._num_iter += 1



import copy
import operator
import os
import numpy as np
import wandb
import sys
from importlib import import_module
import keras
import keras.backend as K


class WandbCallback(keras.callbacks.Callback):
    """WandB Keras Callback.

    Automatically saves history and summary data.  Optionally logs gradients, writes modes,
    and saves example images.

    Optionally saves the best model while training.

    Optionally logs weights and gradients during training.

    """

    def __init__(self, monitor='val_loss', verbose=0, mode='auto',
                 save_weights_only=False, log_weights=False, log_gradients=False,
                 save_model=True, training_data=None, validation_data=[],
                 labels=[], data_type="image"
                 ):
        """Constructor.

        # Arguments
            monitor: quantity to monitor.
            mode: one of {auto, min, max}.
                'min' - save model when monitor is minimized
                'max' - save model when monitor is maximized
                'auto' - try to guess when to save the model
            save_weights_only: if True, then only the model's weights will be
                saved (`model.save_weights(filepath)`), else the full model
                is saved (`model.save(filepath)`).
            save_model:
                True - save a model when monitor beats all previous epochs
                False - don't save models
            log_weights: if True save the weights in wandb.history
            log_gradients: if True log the training gradients in wandb.history
            training_data: tuple (X,y) needed for calculating gradients
            validation_data: numpy array of validation data
            data_type: the type of data we're saving, default "image"
            labels: list of labels
        """
        if wandb.run is None:
            raise wandb.Error(
                'You must call wandb.init() before WandbCallback()')
        self._validation_data = validation_data
        self.labels = labels
        self.data_type = data_type

        self.monitor = monitor
        self.verbose = verbose
        self.save_weights_only = save_weights_only

        self.filepath = os.path.join(wandb.run.dir, 'model-best.h5')
        self.save_model = save_model
        self.log_weights = log_weights
        self.log_gradients = log_gradients
        self.training_data = training_data

        if self.training_data:
            if len(self.training_data) != 2:
                raise ValueError("training data must be a tuple of length two")

        # From Keras
        if mode not in ['auto', 'min', 'max']:
            print('WandbCallback mode %s is unknown, '
                  'fallback to auto mode.' % (mode))
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = operator.lt
            self.best = float('inf')
        elif mode == 'max':
            self.monitor_op = operator.gt
            self.best = float('-inf')
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = operator.gt
                self.best = float('-inf')
            else:
                self.monitor_op = operator.lt
                self.best = float('inf')

    def set_params(self, params):
        self.params = params

    def set_model(self, model):
        self.model = model

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        # history
        row = {'epoch': epoch}
        row.update(logs)

        if self.log_weights:
            weights_metrics = self._log_weights()
            row.update(weights_metrics)

        if self.log_gradients:
            gradients_metrics = self._log_gradients()
            row.update(gradients_metrics)

        if self.data_type == "image" and len(self._validation_data) > 0:
            wandb.run.history.row.update({"examples": self._log_images()})
        wandb.run.history.add(row)

        # summary
        self.current = logs.get(self.monitor)
        if self.current is None:    # validation data wasn't set
            #            print('Can save best model only with %s available, '
            #                  'skipping.' % (self.monitor))
            wandb.run.summary.update(row)
            return

        copied = copy.copy(row)
        if self.monitor_op(self.current, self.best):
            copied.pop('epoch')
            wandb.run.summary.update(copied)
            if self.save_model:
                self._save_model(epoch)

    def on_batch_begin(self, batch, logs=None):
        pass

    def on_batch_end(self, batch, logs=None):
        pass

    def on_train_begin(self, logs=None):
        pass

    def on_train_end(self, logs=None):
        pass

    def _log_images(self):
        indices = np.random.choice(self._validation_data.shape[0], 36)
        test_data = self._validation_data[indices]
        labels = np.argmax(self.model.predict(test_data), axis=1)
        if len(self.labels) > 0:
            captions = []
            for label in labels:
                try:
                    captions.append(self.labels[label])
                except IndexError:
                    captions.append(label)
        else:
            captions = labels
        return [wandb.Image(data, caption=captions[i]) for i, data in enumerate(test_data)]

    def _log_weights(self):
        metrics = {}
        for layer in self.model.layers:
            weights = layer.get_weights()
            if len(weights) == 1:
                metrics[layer.name] = np.mean(weights[0])
            elif len(weights) == 2:
                metrics[layer.name + ".weights-mean"] = np.mean(weights[0])
                metrics[layer.name + ".bias-mean"] = np.mean(weights[1])
        return metrics

    def _log_gradients(self):
        if (not self.training_data):
            raise ValueError(
                "Need to pass in training data if logging gradients")

        X_train = self.training_data[0]
        y_train = self.training_data[1]
        metrics = {}
        weights = self.model.trainable_weights  # weight tensors
        # filter down weights tensors to only ones which are trainable
        weights = [weight for weight in weights
                   if self.model.get_layer(weight.name.split('/')[0]).trainable]

        gradients = self.model.optimizer.get_gradients(
            self.model.total_loss, weights)  # gradient tensors
        input_tensors = [self.model.inputs[0],  # input data
                         # how much to weight each sample by
                         self.model.sample_weights[0],
                         self.model.targets[0],  # labels
                         K.learning_phase(),  # train or test mode
                         ]

        get_gradients = K.function(inputs=input_tensors, outputs=gradients)

        grads = get_gradients([X_train, np.ones(len(y_train)), y_train])

        for (weight, grad) in zip(weights, grads):
            metrics[weight.name.split(':')[0] + ".grad-mean"] = np.mean(grad)
            metrics[weight.name.split(':')[0] + ".grad-stddev"] = np.std(grad)
        return metrics

    def _save_model(self, epoch):
        if self.verbose > 0:
            print('Epoch %05d: %s improved from %0.5f to %0.5f,'
                  ' saving model to %s'
                  % (epoch, self.monitor, self.best,
                     self.current, self.filepath))
        self.best = self.current

        try:
            if self.save_weights_only:
                self.model.save_weights(self.filepath, overwrite=True)
            else:
                self.model.save(self.filepath, overwrite=True)
        except ImportError:
            print("Warning: Can't save model without h5py installed")
            self.save_model = False



import json
import pytest
import os
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
from flask import url_for


def response_json(response):
    return json.loads(response.data.decode())


j = lambda **kwargs: json.dumps(kwargs)


def graphql_url(**url_params):
    string = url_for('graphql.graphql')

    if url_params:
        string += '?' + urlencode(url_params)

    return string


basic_fixture_path = os.path.join(
    os.path.dirname(__file__), "fixtures/basic/wandb")


@pytest.fixture
def app(request):
    marker = request.node.get_marker('base_path')
    path = basic_fixture_path
    if marker:
        path = os.path.join(
            os.path.dirname(__file__), "fixtures", marker.args[0], "wandb")
    from wandb.board.app import create_app
    return create_app("testing", path)



import pytest
import os
import json
import six
import numpy as np
from click.testing import CliRunner

from wandb.history import History
from wandb import media
import torch


@pytest.fixture
def history():
    with CliRunner().isolated_filesystem():
        yield History("wandb-history.jsonl")


def di(row):
    """Returns a dict_items object for easier comparison"""
    return six.viewitems(row)


def disk_history():
    """Reads history from disk and returns an array of dicts"""
    return History("wandb-history.jsonl").rows


def test_history_default(history):
    history.add({"loss": 0.5})
    h = disk_history()
    assert di({"loss": 0.5, "_step": 0}) <= di(h[0])
    assert "_runtime" in h[0].keys()


def test_history_multi_write(history):
    history.row.update({"epoch": 1, "val_loss": 1})
    history.add({"loss": 0.5})
    h = disk_history()
    assert di({"loss": 0.5, "val_loss": 1, "epoch": 1}) <= di(h[0])


def test_history_explicit_write(history):
    history.add({"loss": 0.5})
    history.add({"loss": 0.6})
    h = disk_history()
    assert h[0]["loss"] == 0.5
    assert h[-1]["loss"] == 0.6


def test_step_context(history):
    with history.step() as h:
        h.add({"loss": 0.2})
        h.row["epoch"] = 1
    h = disk_history()
    assert di({"loss": 0.2, "epoch": 1}) <= di(h[0])


def test_step_context_no_compute(history):
    with history.step(compute=False) as h:
        h.add({"loss": 0.2})
        h.row["epoch"] = 1
        if h.compute:
            raise ValueError()
    h = disk_history()
    assert len(h) == 0


def test_step_context_global(history):
    with history.step():
        history.add({"foo": "bar"})
    h = disk_history()
    assert di({"foo": "bar"}) <= di(h[0])


def test_stream_step(history):
    with history.stream("batch").step() as h:
        h.add({"foo": "bar"})
    h = disk_history()
    assert di({"_stream": "batch", "foo": "bar"}) <= di(h[0])


def test_list_of_images(history):
    image = np.random.randint(255, size=(28, 28))
    history.add({"images": [media.Image(image)]})
    h = disk_history()
    assert h[0]["images"] == {'_type': 'images',
                              'count': 1, 'height': 28, 'width': 28}


def test_stream(history):
    history.stream("foo").add({"acc": 1})
    h = disk_history()
    assert di({"_stream": "foo", "acc": 1}) <= di(h[0])


@pytest.mark.skip("This stopped working for reasons not clear to me...")
def test_torch(history):
    with history.step():
        history.torch.log_stats(
            torch.autograd.Variable(torch.randn(
                2, 2).type(torch.FloatTensor), requires_grad=True), "layer1")
    h = disk_history()
    assert "_layer1-0.50" in h[0].keys()


def test_torch_no_compute(history):
    with history.step(False):
        history.torch.log_stats(
            torch.autograd.Variable(torch.randn(
                2, 2).type(torch.FloatTensor), requires_grad=True), "layer1")
    h = disk_history()
    assert len(h) == 0



#!/usr/bin/env python
"""An internal CLI used by the agent and for "headless" mode.
"""

import json
import logging
import os
import socket
import sys
import time

import wandb
import wandb.api
import wandb.run_manager
import wandb.wandb_run


def headless(args):
    """Headless mode is where we start a monitoring / syncing
    process to watch an already-running user process. It's like
    `wandb run` for a user process that has already started.

    The user process that calls this waits for a signal that
    everything is ready, which is sent at the end of rm.wrap_existing_process
    """
    user_process_pid = args['pid']
    stdout_master_fd = args['stdout_master_fd']
    stderr_master_fd = args['stderr_master_fd']

    run = wandb.wandb_run.Run.from_environment_or_defaults()

    api = wandb.api.Api()
    api.set_current_run_id(run.id)

    rm = wandb.run_manager.RunManager(
        api, run, cloud=args['cloud'], job_type=args['job_type'],
        port=args['port'])
    rm.wrap_existing_process(
        user_process_pid, stdout_master_fd, stderr_master_fd)


def agent_run(args):
    """A version of `wandb run` that the agent uses to run things.
    """
    run = wandb.wandb_run.Run.from_environment_or_defaults()

    api = wandb.api.Api()
    api.set_current_run_id(run.id)

    # TODO: better failure handling
    root = api.git.root
    remote_url = api.git.remote_url
    host = socket.gethostname()
    # handle non-git directories
    if not root:
        root = os.path.abspath(os.getcwd())
        remote_url = 'file://%s%s' % (host, root)

    upsert_result = api.upsert_run(id=run.storage_id,
                                   name=run.id,
                                   project=api.settings("project"),
                                   entity=api.settings("entity"),
                                   config=run.config.as_dict(), description=run.description, host=host,
                                   program_path=args['program'], repo=remote_url, sweep_name=run.sweep_id)
    run.storage_id = upsert_result['id']
    env = dict(os.environ)
    run.set_environment(env)

    try:
        rm = wandb.run_manager.RunManager(api, run)
    except wandb.run_manager.Error:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        wandb.termerror('An Exception was raised during setup, see %s for full traceback.' %
                        util.get_log_file_path())
        wandb.termerror(exc_value)
        if 'permission' in str(exc_value):
            wandb.termerror(
                'Are you sure you provided the correct API key to "wandb login"?')
        lines = traceback.format_exception(
            exc_type, exc_value, exc_traceback)
        logger.error('\n'.join(lines))
    else:
        rm.run_user_process(args['program'], args['args'], env)




def main():
    wandb.try_to_set_up_logging()

    args = json.loads(sys.argv[1])
    command = args['command']
    if command == 'headless':
        headless(args)
    elif command == 'agent-run':
        agent_run(args)
    else:
        assert False


if __name__ == '__main__':
    main()



import wandb
import numpy as np
import PIL
import os
from click.testing import CliRunner

image = np.random.randint(255, size=(28, 28))


def test_captions():
    wbone = wandb.Image(image, caption="Cool")
    wbtwo = wandb.Image(image, caption="Nice")
    assert wandb.Image.captions([wbone, wbtwo]) == ["Cool", "Nice"]


def test_transform():
    with CliRunner().isolated_filesystem():
        meta = wandb.Image.transform([wandb.Image(image)], ".", "test.jpg")
        assert meta == {'_type': 'images',
                        'count': 1, 'height': 28, 'width': 28}
        assert os.path.exists("media/images/test.jpg")


def test_guess_mode():
    image = np.random.randint(255, size=(28, 28, 3))
    wbimg = wandb.Image(image)
    assert wbimg.image.mode == "RGB"


def test_pil():
    pil = PIL.Image.new("L", (28, 28))
    img = wandb.Image(pil)
    assert img.image == pil



#!/usr/bin/env python

from __future__ import print_function

import datetime
import logging
import os
import re
import time

from six import b

logger = logging.getLogger(__name__)


class LineBuffer(object):
    """Streaming string parser that extracts lines."""

    def __init__(self):
        self._buf = []
        # If the backend receive a line ending with \r, it assumes that
        # line is part of a progress bar and ignores it. To handle terminal
        # output properly on Windows we need to ensure we always pass along
        # a newline if it comes after a carriage return.
        self._line_end_re = re.compile(b('\r\n|\r|\n'))

    def add_string(self, data):
        """Process some data splitting it into complete lines and buffering the rest

        Args:
            data: A `str` in Python 2 or `bytes` in Python 3
        Returns:
            list of complete lines ending with a carriage return (eg. a progress
            bar) or a newline.
        """
        lines = []
        while data:
            match = self._line_end_re.search(data)
            if match is None:
                chunk = data
            else:
                chunk = data[:match.end()]

            data = data[len(chunk):]

            if self._buf and self._buf[-1].endswith(b('\r')) and not chunk.startswith(b('\n')):
                # if we get a carriage return followed by something other than
                # a newline then we assume that we're overwriting the current
                # line (ie. a progress bar)
                #
                # We don't terminate lines that end with a carriage return until
                # we see what's coming next so we can distinguish between a
                # progress bar situation and a Windows line terminator.
                #
                # TODO(adrian): some day these hacks should be replaced with
                # real terminal emulation
                lines.append(self._finish_line())

            self._buf.append(chunk)
            if chunk.endswith(b('\n')):
                lines.append(self._finish_line())

        return lines

    def _finish_line(self):
        line = b('').join(self._buf).decode('utf-8')
        self._buf = []
        return line


class TextStreamPusher(object):
    """Pushes a stream of text, line by line, to wandb."""

    def __init__(self, fsapi, filename, line_prepend='', prepend_timestamp=False, lock_function=None):
        """Constructor.

        Args:
            fsapi: api.FileStreamApi instance
            filename: Name of the file this stream is pushed to.
            line_prepend: string to prepend to every line for this stream.
            prepend_timestamp: If true a timestamp will be prepended to each line
                (after line_prepend).
            lock_function: Optional function to call before writing data to the
                backend. We use this to block writes while the backend is down.
                The function should both acquire and release the lock because we
                don't actually want to hold the lock.
        """
        self._fsapi = fsapi
        self._filename = filename
        if line_prepend:
            line_prepend += ' '
        self._line_prepend = line_prepend
        self._prepend_timestamp = prepend_timestamp
        self._lock_function = lock_function
        self._line_buffer = LineBuffer()

    def write_string(self, message, cur_time=None):
        return self.write(message.encode('utf-8'), cur_time)

    def write(self, message, cur_time=None):
        """Write some text to the pusher.

        Args:
            message: a string to push for this file.
            cur_time: used for unit testing. override line timestamp.
        """
        #print('ts write', repr(message))
        if self._lock_function:
            self._lock_function()
        if cur_time is None:
            cur_time = time.time()
        lines = self._line_buffer.add_string(message)
        for line in lines:
            #print('ts line', repr(line))
            timestamp = ''
            if self._prepend_timestamp:
                timestamp = datetime.datetime.utcfromtimestamp(
                    cur_time).isoformat() + ' '
            line = '%s%s%s' % (self._line_prepend, timestamp, line)
            self._fsapi.push(self._filename, line)

    def close(self):
        """Close the file."""
        # Force a final line to clear whatever might be in the buffer.
        self.write_string(os.linesep)



from flask import Blueprint
import flask_cors
import logging
import traceback
from flask_graphql import GraphQLView
from wandb.board.app.graphql.schema import schema
from wandb.board.app.util.errors import format_error

logger = logging.getLogger(__name__)

graphql = Blueprint('graphql', __name__)
flask_cors.CORS(graphql)
GraphQLView.format_error = format_error


graphql.add_url_rule(
    '/graphql', view_func=GraphQLView.as_view('graphql', schema=schema, graphiql=True))



#!/usr/bin/env python

from __future__ import print_function

import collections
import contextlib
import copy
import json
import os
import time
from threading import Lock
import warnings
import weakref
import six

from wandb.wandb_torch import TorchHistory
import wandb
from wandb import util
from wandb import media


class History(object):
    """Used to store data that changes over time during runs. """

    def __init__(self, fname, out_dir='.', add_callback=None, stream_name="default"):
        self._start_time = wandb.START_TIME
        self.out_dir = out_dir
        self.fname = os.path.join(out_dir, fname)
        self.rows = []
        self.row = {}
        self.stream_name = stream_name
        # during a batched context logging may still be disabled. we do it this way
        # so people don't have to litter their code with conditionals
        self.compute = False
        self.batched = False
        # not all rows have the same keys. this is the union of them all.
        self._keys = set()
        self._process = "user" if os.getenv("WANDB_INITED") else "wandb"
        self._streams = {}
        self._steps = 0
        self._lock = Lock()
        self._torch = None
        self.load()
        self._file = open(self.fname, 'a')
        self._add_callback = add_callback

    def load(self):
        self.rows = []
        try:
            # only preload the default stream, TODO: better stream support
            if self.stream_name == "default":
                with open(self.fname) as f:
                    for line in f:
                        try:
                            self._index(json.loads(line))
                        except TypeError:
                            print('warning: malformed history line: %s...' %
                                  line[:40])
                # initialize steps and run time based on existing data.
                if '_step' in self.row:
                    self._steps = self.row['_step'] + 1
                # fudge the start_time to compensate for previous run length
                if '_runtime' in self.row:
                    self._start_time = wandb.START_TIME - self.row['_runtime']
                self.row = {}
        except IOError:
            pass

    def keys(self):
        media_keys = []
        if self.rows:
            media_keys = [k for k, v in six.iteritems(
                self.rows[-1]) if isinstance(v, dict) and v.get("_type")]
        return [k for k in self._keys - set(media_keys) if not k.startswith("_")]

    def stream(self, name):
        """stream can be used to record different time series:

        run.history.stream("batch").add({"gradients": 1})
        """
        if self.stream_name != "default":
            raise ValueError("Nested streams aren't supported")
        if self._streams.get(name) == None:
            self._streams[name] = History(self.fname, out_dir=self.out_dir,
                                          add_callback=self._add_callback, stream_name=name)
        return self._streams[name]

    def column(self, key):
        """Iterator over a given column, skipping rows that don't have a key
        """
        for row in self.rows:
            if key in row:
                yield row[key]

    def add(self, row={}):
        """Adds keys to history and writes the row.  If row isn't specified, will write
        the current state of row.

        run.history.row["duration"] = 1.0
        run.history.add({"loss": 1})
        => {"duration": 1.0, "loss": 1}

        """
        if not isinstance(row, collections.Mapping):
            raise wandb.Error('history.add expects dict-like object')
        self.row.update(row)
        if not self.batched:
            self._write()

    @contextlib.contextmanager
    def step(self, compute=True):
        """Context manager to gradually build a history row, then commit it at the end.

        To reduce the number of conditionals needed, code can check run.history.compute:

        with run.history.step(batch_idx % log_interval == 0):
            run.history.add({"nice": "ok"})
            if run.history.compute:
                # Something expensive here
        """
        self.row = {}
        self.batched = True
        self.compute = compute
        yield self
        if compute:
            self._write()

    @property
    def torch(self):
        if self._torch is None:
            self._torch = TorchHistory(self)
        return self._torch

    def _index(self, row):
        """Internal row adding method that updates step, and keys"""
        self.row = row
        self.rows.append(row)
        self._keys.update(row.keys())
        self._steps += 1

    def _transform(self):
        """Transforms media classes into the proper format before writing"""
        for key, val in six.iteritems(self.row):
            if type(val) in (list, tuple) and len(val) > 0:
                is_image = [isinstance(v, media.Image) for v in val]
                if all(is_image):
                    self.row[key] = media.Image.transform(val, self.out_dir,
                                                          "{}_{}.jpg".format(key, self.row["_step"]))
                elif any(is_image):
                    raise ValueError(
                        "Mixed media types in the same list aren't supported")

    def _write(self):
        if self.row:
            self._lock.acquire()
            try:
                self.row['_runtime'] = time.time() - self._start_time
                self.row['_timestamp'] = time.time()
                self.row['_step'] = self._steps
                if self.stream_name != "default":
                    self.row["_stream"] = self.stream_name
                self._transform()
                self._file.write(util.json_dumps_safer(self.row))
                self._file.write('\n')
                self._file.flush()
                self._index(self.row)
                if self._add_callback:
                    self._add_callback(self.row)
                self.row = {}
            finally:
                self._lock.release()
                return True
        else:
            return False

    def close(self):
        self._write()
        self._lock.acquire()
        try:
            self._file.close()
            self._file = None
        finally:
            self._lock.release()



import os


class Config:
    BOARD_ADMIN = os.environ.get('BOARD_ADMIN')
    SSL_REDIRECT = False
    #SERVER_NAME = "localhost:7177"

    @staticmethod
    def init_app(app):
        pass


class DevelopmentConfig(Config):
    DEBUG = True


class TestingConfig(Config):
    TESTING = True


class ProductionConfig(Config):
    pass


config = {
    'development': DevelopmentConfig,
    'testing': TestingConfig,
    'production': ProductionConfig,
    'default': DevelopmentConfig
}



import json
import os
import threading
import time
import socket
import getpass
from datetime import datetime

from wandb import util
import wandb

METADATA_FNAME = 'wandb-metadata.json'


class Meta(object):
    """Used to store metadata during and after a run."""

    HEARTBEAT_INTERVAL_SECONDS = 15

    def __init__(self, api, out_dir='.'):
        self.fname = os.path.join(out_dir, METADATA_FNAME)
        self._api = api
        self._shutdown = False
        try:
            self.data = json.load(open(self.fname))
        except (IOError, ValueError):
            self.data = {}
        self.lock = threading.Lock()
        self.setup()
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True
        self._thread.start()

    def setup(self):
        self.data["root"] = os.getcwd()
        if self._api.git.enabled:
            self.data["git"] = {
                "remote": self._api.git.remote_url,
                "commit": self._api.git.last_commit
            }
            self.data["email"] = self._api.git.email
            self.data["root"] = self._api.git.root or self.data["root"]

        self.data["startedAt"] = datetime.utcfromtimestamp(
            wandb.START_TIME).isoformat()
        self.data["host"] = socket.gethostname()
        self.data["username"] = os.getenv("WANDB_USERNAME", getpass.getuser())
        try:
            import __main__
            self.data["program"] = __main__.__file__
        except (ImportError, AttributeError):
            self.data["program"] = '<python with no main file>'
        self.data["state"] = "running"
        self.write()

    def write(self):
        self.lock.acquire()
        try:
            self.data["heartbeatAt"] = datetime.utcnow().isoformat()
            with open(self.fname, 'w') as f:
                s = util.json_dumps_safer(self.data, indent=4)
                f.write(s)
                f.write('\n')
        finally:
            self.lock.release()

    def shutdown(self):
        self._shutdown = True
        self._thread.join()

    def _thread_body(self):
        seconds = 0
        while True:
            if seconds > self.HEARTBEAT_INTERVAL_SECONDS or self._shutdown:
                self.write()
                seconds = 0
            if self._shutdown:
                break
            else:
                time.sleep(2)
                seconds += 2



import wandb
from wandb import wandb_run
from wandb.keras import WandbCallback
import pytest
from click.testing import CliRunner
import os
import json
from .utils import git_repo
from keras.layers import Dense
from keras.models import Sequential
import sys


@pytest.fixture
def dummy_model():
    model = Sequential()
    model.add(Dense(32, activation='relu', input_dim=10))
    model.add(Dense(1, activation='sigmoid'))
    model.compile(optimizer='rmsprop',
                  loss='binary_crossentropy',
                  metrics=['accuracy'])
    return model


@pytest.fixture
def dummy_data():
    import numpy as np
    data = np.random.random((100, 10))
    labels = np.random.randint(2, size=(100, 1))
    return (data, labels)


def test_basic_keras(dummy_model, dummy_data, git_repo):
    wandb.run = wandb_run.Run.from_environment_or_defaults()
    dummy_model.fit(*dummy_data, epochs=2, batch_size=36,
                    callbacks=[WandbCallback()])
    assert wandb.run.summary["epoch"] == 1



import os
from wandb import util


class Media(object):
    @classmethod
    def from_type(cls, type):
        pass


class Image(object):
    def __init__(self, data, mode=None, caption=None):
        """
        Accepts numpy array of image data, or a PIL image. The class attempts to infer
        the data format and converts it.
        """
        try:
            from PIL import Image as PILImage
        except ImportError:
            raise ValueError(
                "wandb.Image requires the PIL package: pip install pillow")
        if type(data) == PILImage.Image:
            self.image = data
        else:
            self.image = PILImage.fromarray(
                self.to_uint8(data), mode=mode or self.guess_mode(data))
        self.caption = caption

    def guess_mode(self, data):
        """
        Guess what type of image the np.array is representing 
        """
        # TODO: do we want to support dimensions being at the beginning of the array?
        if data.shape[-1] == 1 or len(data.shape) == 2:
            return "L"
        elif data.shape[-1] == 3:
            return "RGB"
        elif data.shape[-1] == 4:
            return "RGBA"
        else:
            raise ValueError(
                "Un-supported shape for image conversion %s" % list(data.shape))

    def to_uint8(self, data):
        """
        Converts floating point image on the range [0,1] and integer images
        on the range [0,255] to uint8, clipping if necessary.
        """
        try:
            import numpy as np
        except ImportError:
            raise ValueError(
                "wandb.Image requires numpy if not supplying PIL Images: pip install numpy")
        if issubclass(data.dtype.type, np.floating):
            data = (data * 255).astype(np.int32)
        assert issubclass(data.dtype.type, np.integer), 'Illegal image format.'
        return data.clip(0, 255).astype(np.uint8)

    @staticmethod
    def transform(images, out_dir, fname):
        """
        Combines a list of images into a single sprite returning meta information
        """
        from PIL import Image as PILImage
        base = os.path.join(out_dir, "media", "images")
        width, height = images[0].image.size
        sprite = PILImage.new(
            mode='RGB',
            size=(width * len(images), height),
            color=(0, 0, 0, 0))
        for i, image in enumerate(images):
            location = width * i
            sprite.paste(image.image, (location, 0))
        util.mkdir_exists_ok(base)
        sprite.save(os.path.join(base, fname), transparency=0)
        meta = {"width": width, "height": height,
                "count": len(images), "_type": "images"}
        captions = Image.captions(images)
        if captions:
            meta["captions"] = captions
        return meta

    @staticmethod
    def captions(images):
        if images[0].caption:
            return [i.caption for i in images]
        else:
            return False



import logging

import graphene
from graphene import relay
from .loader import data, find_run, settings
from wandb.board.app.models import History, Events, Log
from wandb.board.app.util.errors import NotFoundError
from datetime import datetime

logger = logging.getLogger(__name__)

# Missing: state, commit, runTime, job


class UserType(graphene.ObjectType):
    id = graphene.ID(required=True)
    email = graphene.String()
    username = graphene.String()
    photoUrl = graphene.String()
    admin = graphene.String()
    defaultFramework = graphene.String()

    def resolve_id(self, info, **args):
        return "id"


class LogLine(graphene.ObjectType):
    line = graphene.String()
    number = graphene.Int()
    level = graphene.String()
    id = graphene.String()


class LogLineConnection(relay.Connection):
    class Meta:
        node = LogLine


class SweepType(graphene.ObjectType):
    id = graphene.String()
    name = graphene.String()
    createdAt = graphene.String()
    updatedAt = graphene.String()
    heartbeatAt = graphene.String()
    description = graphene.String()
    runCount = graphene.Int()
    state = graphene.String()
    user = graphene.Field(UserType)


class SweepConnectionType(relay.Connection):
    class Meta:
        node = SweepType


class FileType(graphene.ObjectType):
    id = graphene.String()
    name = graphene.String()
    url = graphene.String(upload=graphene.Boolean())
    sizeBytes = graphene.Int()
    updatedAt = graphene.String()


class FileConnectionType(relay.Connection):
    class Meta:
        node = FileType


class Run(graphene.ObjectType):
    class Meta:
        interfaces = (relay.Node, )

    id = graphene.ID(required=True)
    name = graphene.String()
    path = graphene.String()
    host = graphene.String()
    createdAt = graphene.String()
    description = graphene.String()
    github = graphene.String()
    commit = graphene.String()
    state = graphene.String()
    patch = graphene.String()
    config = graphene.types.json.JSONString()
    summaryMetrics = graphene.types.json.JSONString()
    systemMetrics = graphene.types.json.JSONString()
    requestSubscribe = graphene.Boolean()
    heartbeatAt = graphene.String()
    events = graphene.List(graphene.String)
    tags = graphene.List(graphene.String)
    history = graphene.List(graphene.String, samples=graphene.Int())
    logLines = relay.ConnectionField(
        LogLineConnection)

    framework = graphene.String()
    shouldStop = graphene.Boolean()
    sweep = graphene.Field(SweepType)
    fileCount = graphene.Int()
    exampleTableColumns = graphene.types.json.JSONString()
    exampleTableTypes = graphene.types.json.JSONString()
    exampleTable = graphene.types.json.JSONString()
    files = relay.ConnectionField(FileConnectionType)

    user = graphene.Field(UserType)

    def resolve_files(self, info, **args):
        return []

    def resolve_tags(self, info, **args):
        return []

    def resolve_fileCount(self, info, **args):
        return 0

    def resolve_name(self, info, **args):
        return self.id

    def resolve_history(self, info, **args):
        return filter(None, History(self.path).read().split("\n"))

    def resolve_events(self, info, **args):
        return filter(None, Events(self.path).read().split("\n"))

    def resolve_logLines(self, info, **args):
        return Log(self.path).lines()

    def resolve_exampleTable(self, info, **args):
        return ""

    def __repr__(self):
        return "<Run %s>" % self.id


class BucketType(Run):
    pass


class RunConnection(relay.Connection):
    class Meta:
        node = Run


class BucketConnection(relay.Connection):
    class Meta:
        node = BucketType


class Project(graphene.ObjectType):
    id = graphene.ID(required=True)
    name = graphene.String()
    access = graphene.String()
    entityName = graphene.String()
    description = graphene.String()
    createdAt = graphene.String()
    summaryMetrics = graphene.String()
    views = graphene.JSONString()
    runCount = graphene.Int()
    bucketCount = graphene.Int()
    requestSubscribe = graphene.Boolean()

    runs = relay.ConnectionField(
        RunConnection, entityName=graphene.String(), names=graphene.List(graphene.String),
        jobKey=graphene.String(), order=graphene.String())
    buckets = relay.ConnectionField(
        BucketConnection, entityName=graphene.String(), names=graphene.List(graphene.String),
        jobKey=graphene.String(), order=graphene.String())

    run = graphene.Field(Run, name=graphene.String())
    bucket = graphene.Field(BucketType, name=graphene.String())
    sweeps = relay.ConnectionField(SweepConnectionType)

    def resolve_description(self, info, **args):
        return settings.get_project("description")

    def resolve_sweeps(self, info, **args):
        return []

    def resolve_run(self, info, **args):
        return find_run(args["name"])

    def resolve_bucket(self, info, **args):
        return find_run(args["name"])

    def resolve_entityName(self, info, **args):
        return "board"

    def resolve_runs(self, info, **args):
        return data["Runs"]

    def resolve_buckets(self, info, **args):
        return data["Runs"]

    def resolve_id(self, info, **args):
        return "default"

    def resolve_name(self, info, **args):
        return "default"

    def resolve_summaryMetrics(self, info, **args):
        return "{}"

    def resolve_views(self, info, **args):
        return settings.get_project("views")


class ProjectConnection(relay.Connection):
    class Meta:
        node = Project


class ModelType(Project):
    pass


class ModelConnection(relay.Connection):
    class Meta:
        node = ModelType


class Query(graphene.ObjectType):
    project = graphene.Field(
        Project, name=graphene.String(), entityName=graphene.String())
    model = graphene.Field(
        ModelType, name=graphene.String(), entityName=graphene.String())
    models = relay.ConnectionField(
        ModelConnection, entityName=graphene.String())
    viewer = graphene.Field(
        UserType
    )

    def resolve_project(self, info, **args):
        return Project()

    def resolve_model(self, info, **args):
        return ModelType()

    def resolve_models(self, info, **args):
        return [ModelType()]

    def resolve_viewer(self, info, **args):
        return UserType()


class UpsertProject(relay.ClientIDMutation):
    class Input:
        name = graphene.String()
        description = graphene.String()
        id = graphene.String()
        framework = graphene.String()
        access = graphene.String()
        entityName = graphene.String()
        repo = graphene.String()
        access = graphene.String()
        views = graphene.JSONString()

    model = graphene.Field(ModelType)
    entityName = graphene.String()
    inserted = graphene.Boolean()

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        if input.get("description"):
            settings.set_project("description", input["description"])
        if input.get("views"):
            settings.set_project("views", input["views"])

        settings.save()
        return UpsertProject(model=Project(), entityName="board", inserted=False)


class UpsertRun(relay.ClientIDMutation):
    class Input:
        name = graphene.String()
        description = graphene.String()
        sweep = graphene.String()
        id = graphene.String()
        framework = graphene.String()
        config = graphene.types.json.JSONString()
        commit = graphene.String()
        host = graphene.String()
        debug = graphene.Boolean()
        entityName = graphene.String()
        modelName = graphene.String()
        jobProgram = graphene.String()
        jobRepo = graphene.String()
        jobType = graphene.String()

    bucket = graphene.Field(BucketType)
    inserted = graphene.Boolean()

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        run = find_run(input["id"], mutator=True)
        if input.get("description"):
            run.description = input["description"]
        return UpsertRun(bucket=find_run(input["id"]), inserted=False)


class Mutation(graphene.ObjectType):
    upsertModel = UpsertProject.Field()
    upsertProject = UpsertProject.Field()
    upsertBucket = UpsertRun.Field()
    upsertRun = UpsertRun.Field()


schema = graphene.Schema(query=Query, mutation=Mutation, types=[Project, Run])



from .utils import runner
import os
import sh
import glob
import json
import time
import signal

train_py = open(os.path.join(os.path.dirname(
    __file__), "fixtures/train.py")).read()


def test_dry_run(runner):
    print("OH NO", dict(os.environ))
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py)
        os.environ["WANDB_MODE"] = "dryrun"
        res = sh.python("train.py")
        run_dir = glob.glob("wandb/dry*")[0]
        assert "loss:" in res
        meta = json.loads(open(run_dir + "/wandb-metadata.json").read())
        assert meta["state"] == "finished"
        assert meta["program"] == "train.py"
        assert meta["exitcode"] == 0
        assert os.path.exists(run_dir + "/output.log")
        assert os.path.exists(run_dir + "/wandb-history.jsonl")
        assert os.path.exists(run_dir + "/wandb-events.jsonl")
        assert os.path.exists(run_dir + "/wandb-summary.json")
        del os.environ["WANDB_MODE"]


def test_dry_run_custom_dir(runner):
    with runner.isolated_filesystem():
        os.environ["WANDB_DIR"] = "/tmp"
        os.environ["WANDB_MODE"] = "dryrun"
        try:
            with open("train.py", "w") as f:
                f.write(train_py)
            res = sh.python("train.py")
            print(res)
            run_dir = glob.glob("/tmp/wandb/dry*")[0]
            assert os.path.exists(run_dir + "/output.log")
        finally:  # avoid stepping on other tests, even if this one fails
            del os.environ["WANDB_DIR"]
            del os.environ["WANDB_MODE"]
            # TODO: this isn't working
            sh.rm("-rf", "/tmp/wandb")


def test_dry_run_exc(runner):
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py.replace("#raise", "raise"))
        os.environ["WANDB_MODE"] = "dryrun"
        try:
            res = sh.python("train.py")
        except sh.ErrorReturnCode as e:
            res = e.stdout
        print(res)
        run_dir = glob.glob("wandb/dry*")[0]
        meta = json.loads(open(run_dir + "/wandb-metadata.json").read())
        assert meta["state"] == "failed"
        assert meta["exitcode"] == 1
        del os.environ["WANDB_MODE"]


def test_dry_run_kill(runner):
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py.replace("#os.kill", "os.kill"))
        os.environ["WANDB_MODE"] = "dryrun"
        res = sh.python("train.py", _bg=True)
        try:
            res.wait()
            print(res)
        except sh.ErrorReturnCode:
            pass
        dirs = glob.glob("wandb/dry*")
        print(dirs)
        run_dir = dirs[0]
        meta = json.loads(open(run_dir + "/wandb-metadata.json").read())
        assert meta["state"] == "killed"
        assert meta["exitcode"] == 255
        del os.environ["WANDB_MODE"]

# TODO: test server communication
